--==============================================================================
--Top level module to test and demonstrate the XSA USB Interface.
--This module demonstates all of the USB operations available in the visual basic FPGA program
------------------------------------------------------------------
--Copywrite Johns Hopkins University ECE department
--This software may be freely used and modified as long as this header is retained. 
------------------------------------------------------------------
--Component USBF_IFC Implements USB enumeration and communication exchanges with
--  the usb port through the USB_DRVR traneciver component.  
--Project reset is held until interface indicates configuration done.
--===========================================================================
library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.numeric_std.all;
use WORK.sdramj.all;                    --modified xess sdramcntl package for XSA brds.
--use work.RJDeclares.all;
use WORK.USBF_Declares.all;             --constants and user type declarations for USB_IFC

----------------------------------------------------------------------
entity USB_Demo is                      --Interact with the Visual Basic FPGA program
  port(
    ----XSTEND BOARD CONNECTIONS        ----------------------------------------
    --clk50:    in              std_logic;  --50 MHz clock for the sdram interface
    clk100 : in  std_logic;             --100 MHz clock for the sdram interface DLL
    fceb   : out std_logic;             --flash not-select 
    swrst  : in  std_logic;             --input from XST reset push button
    --sw1:              in      std_logic;  --pushbutton switch 1

    ----USB CONNECTIONS (with one 1.5K pull up to 3.3v)  ---------------
    dplus  : inout std_logic;           --D+ data line , pin D16   
    dminus : inout std_logic;           --D- data line, pin E14

    --------SDRAM I/O connections       ---------------- 
    sclkfb     : in    std_logic;       -- feedback SDRAM clock with PCB delays
    sclk       : out   std_logic;       -- clock to SDRAM
    cke        : out   std_logic;       -- SDRAM clock-enable
    cs_n       : out   std_logic;       -- SDRAM chip-select
    ras_n      : out   std_logic;       -- SDRAM RAS
    cas_n      : out   std_logic;       -- SDRAM CAS
    we_n       : out   std_logic;       -- SDRAM write-enable
    ba         : out   std_logic_vector( 1 downto 0);  -- SDRAM bank-address
    saddr      : out   std_logic_vector(12 downto 0);  -- SDRAM address bus
    sdat       : inout std_logic_vector(15 downto 0);  --data bus to SDRAM
    dqmh       : out   std_logic;       -- SDRAM DQMH
    dqml       : out   std_logic;       -- SDRAM DQML
    --------------------------------------------------------------
    ----DEBUG ONLY PORTS                ------------------------------------------------
    --sigout1:  out     std_logic;      -- debug lines for scoping- fpga pin E15
    --sigout2:  out     std_logic;      --fpga pin F4
    --sigout3:  out     std_logic;
    lsdp, rsdp : out   std_logic;       --decimal pt
    ls         : out   std_logic_vector(6 downto 0);  --XST segmented LEDs
    rs         : out   std_logic_vector(6 downto 0);
    bg         : out   std_logic_vector(7 downto 0);  --XST bar graph LED
    --bg7:              out     std_logic;
    bg8        : out   std_logic;       --dll lock
    bg9        : out   std_logic        --config done
    );
end USB_Demo;
----------------------------------------------------------------------

architecture arch of USB_Demo is
--Note that all signals associated with the dram interface are declared in the sdramj pkg.
--All constants and types needed for the USB interface are declared in the usb_declares pkg.
--SIGNALS                               -----------------------------
  signal ctr1   : integer range 0 to 50_000_000;  --counter
  signal ctr2   : unsigned(15 downto 0);  --counter
  signal resetp : std_logic;            --project reset waits for usb to configure
  signal buf_ct : integer range 0 to 127;  --pointer for the dram in/out buffers.
  signal debug  : std_logic_vector(64 downto 51);  --*****DEBUG BITS from interface

-- USB signals                          -----------------------------
  signal new_dat     : std_logic;       -- USB new data flag
  signal rd_req      : std_logic;       -- USB read request flag
  signal mode        : std_logic_vector(7 downto 0);  -- USB set_feature data byte
  signal PCIn        : EIGHT_BYTES;     --Input to USB interface component for sending to PC host
  signal mempacket   : EIGHT_BYTES;     --packet to go out from memory reads
  signal datpacket   : EIGHT_BYTES;     --packet to go out when memory not reading
  signal PCOut       : EIGHT_BYTES;     --Input from PC is output from USB interface
--PKT_BYTES and EIGTH_BYTES are identical arrays in this version of the USB interface
  signal config_done : std_logic;

-----------------------------------------------
------Memory read/write signals         ---------------
  type SM_type is (IDLE, GETBYTES, WRITEP, WRPAUSE, CHKFLG, SENDBYTES, RDPAUSE1, RDPAUSE2);
  signal                           state : SM_type;  --dram access state machine
  attribute init                         : string;
  attribute init of state                : signal is "IDLE";  --we powerup in IDLE state
  signal                           addr  : unsigned(HADDR_WIDTH-1 downto 0);  -- host side address generated by state mach.

  signal sm_clk, synced50M, synced100M : std_logic;  --synced 50Mhz and 100Mhz output from dram DLL. 

--------------------------------------
  function LEDHi (value : std_logic_vector(3 downto 0)) return std_logic_vector is
--7-segment active HI LED decode function.
--Assumes data is 4 bits, returns 7 bit LED vector from look-up.
--variable LED: std_logic_vector(6 downto 0);
----LUT implementation in ROM (this caused reg elimination problem sometimes)
    type ARR16X7 is array (0 to 15) of std_logic_vector (6 downto 0);
    constant LEDLUT_H   : ARR16X7 := ("1110111", "0010010", "1011101", "1011011", "0111010", "1101011", "1101111",
                                                            "1010010", "1111111", "1111010", "1111110", "0101111", "1100101", "0011111", "1101101", "1101100");
--LUT for active HI LED's               ------------------------------ 
  begin
    return LEDLUT_H(TO_INTEGER(unsigned(value)) );
  end LEDHi;
--------------------------------------

--=======Set XSA Board Type===================
--------------------------------------
-- select one of two integers based on s true or false
  function brd_select(s : in boolean; a : in integer; b : in integer) return integer is
  begin
    if s then
      return a;
    else
      return b;
    end if;
  end function brd_select;
---------------------------------------
  constant XSABRD       :    integer := 200;  --same values apply to 3S1000 which has same DRAM as XSA200
----Set generics for SDRAM from Board type
  constant FREQ         :    integer := brd_select(XSABRD = 200, 100_000, 50_000);
  constant CLK_DIV      :    real    := real_select(XSABRD = 200, 2.0, 1.0);
  constant NROWS        :    integer := brd_select(XSABRD = 200, 8192, 4096);
  constant NCOLS        :    integer := brd_select(XSABRD = 50, 256, 512);
----==========================================


begin

--===============================================
-- Main USB interface Component.
  UUSB : USBF_IFC
    port map(
      debug       => debug,             --****debug bits out
      clk100M     => synced100M,        --synced 100MHz master clock from DRAM DLL
      --clk100M=>       clk100,         -- external 100MHz master clock
      reset       => resetp,            -- XST reset push button
      PCIn        => PCIn,              --data array to PC host 
      PCOut       => PCOut,             --data array from PC host
      new_dat     => new_dat,           --pulses high when new data packet arrives
      rd_req      => rd_req,            --pulses high when new read packet starts
      mode        => mode,              --mode byte received by set_feature packet.
      ----USB CONNECTIONS               ---------
      dplus       => dplus,             --differential data
      dminus      => dminus,
      ----------------------------
      config_done => config_done
      );
--debug1<= masterclk;
---=============================================

-- shift register based synchronizer to debounce sw1- 3 samples at 33.3 hz/w filtering
--USYNC1: sync_ext generic map(DIV=>1_500_000, SRLEN=>3, INIT=>'0')
-- port map(sig=>not(sw1), reset=>reset, clk=>clkext, sdb=>synced_sw1);  --sync to it.

--=============================================
--PORT CONNECTIONS
--=============================================
-- XSA board connections
  fceb   <= '1';                        -- disable XSA flash memory
--USB waits for DRAM DLL lock to start enumeration because it uses masterclk.
--resetusb<= '1' when lock = '0' else not(swrst);  --usb reset tied to pushbutton after lock
--project resetp waits for successful enumeration to release control to swrst
  resetp <= '1' when lock = '0' else not(swrst);

----===DEBUG===================================
-- put debug signals on leds.
  rs <= LedHi(mode(3 downto 0));        --display the feature mode
--rs<= "0000000";
--show the address activity when reading or writing, else the DDS control byte                          
  ls <= haddr(6 downto 0) when unsigned(mode) > 0 else
-- "0000000";
        LEDHi(debug(64 downto 61));     -- usb_addr
--        PCOutput(2)(6 downto 0);      --show what would normally be the DDS freq control byte.
--ls<=LEDHi(debug(50 downto 47));       --last sent pid
--rs<=LEDHi(debug(12 downto 9));        -- last received pid
--ls<=LEDHi(debug(24 downto 21));       --last recvd addr

--debug(50:1) comes from the DRVR.
--debug(32:1) is token_in, the rest are commented below.
--debug states:40=rcvd ACK, 41=ZLDP sent for status, 42=crc16 failed
--43=got a setup token/in idle, 44=sent an ack, 45=ZLDP recvd, 46=sent a data pkt
--39=got pkt w/wrong addr
--51=rcvd but intrpt,52=devdesc req, 53 configdesc req, 54=status req,
--55=short pkt sent, 56=IN came after all data sent, 57=got ack intrpt, 58=in stall
--59=usb addr rcvd.
--bg(4 downto 0)<= debug(32 downto 28);  --last crc5 received
  bg(0) <= debug(60);                   --a ZLP IN request
--bg(1)<= debug(42);                    --42=crc16-in failed/44=sent an ACK
  bg(1) <= debug(51);                   --recvd a bus reset
--bg(2)<= debug(45);                    --a 0-length OUT pkt received
  bg(2) <= debug(59);                   --recvd a usb addr
  bg(3) <= debug(54);                   --54=status requested
  bg(4) <= debug(57);                   --ack interrupt rcvd    
  bg(5) <= debug(53);                   --53=config desc req
  bg(6) <= debug(52);                   --dev_descr requested after address set
  bg(7) <= debug(55);                   --got down to a short pkt
--sigout1<= debug(33);                  --bit clk
--sigout2<= debug(34);                  --one of the data in or out.
  rsdp  <= debug(56);                   --IN req error(negative bytes to send).
  lsdp  <= debug(58);                   --in stall condition.
--lsdp<= debug(42);                     --stall pkt was sent
--lsdp<= PCOutput(1)(0);                --Display what would normally be the DDS on-off control
-- Bottom most bg LEDs turn on when we are locked and USB configured
  bg8   <= lock;                        --indicate that DRAM DLL's locked. clk1x now available
  bg9   <= config_done;                 --indicates enumeration completed successfully. USB can now operate.
--========================================================

---Procedure to test writes, reads, and continuous reads  -----
--When a read is performed by the PC, the values in array PCInput are sent in a packet.
--When a write is performed by the PC, the array PCOutput latches the data until the next write.
  PCT : process (resetp, synced50M, ctr2, mode, PCOut) is
  begin
--generate counter state to send up to PC
    if resetp = '1' then                --resetp waits for configuration
      ctr1         <= 0;
      ctr2         <= (others => '0');
    elsif rising_edge(synced50M) then   --count all received data packets
                                        --clkext<= not clkext;
      if ctr1 < 3_125_000 then          --this is to generate a 16hz counter from masterclk
        ctr1       <= ctr1 + 1;
      else
        ctr1       <= 0;
        ctr2       <= ctr2 + 1;
      end if;
    end if;
--Data for read reports back to PC in mode 0 or 1.
--In mode 2 the data comes from the read Mem state machine
    for i in 1 to 5 loop                --first 5 are just what was last sent by PC
      datpacket(i) <= PCOut(i);
    end loop;
    datpacket(6)   <= std_logic_vector(ctr2(15 downto 8));  --16 hz counter state
    datpacket(7)   <= std_logic_vector(ctr2(7 downto 0));
    datpacket(8)   <= mode;             --current mode byte
------------------------------------------
  end process;
  PCIn             <= datpacket when unsigned(mode) = 0 else mempacket;  --mode=2 is mem read mode

---===Test of USB Memory Loads==================================================
--This is a sample code to test loading and dumping the XSA dram via the usb interface.
--We instantiate a dram interface, and then load and read the memory with a state machine
-- that is compatible with the functions Load_MemUSB and Read_MemUSB in the
-- visual basic communication program on the PC side of the USB transactions.
---*****SDRAM control interface********************************

  rst_int <= '1' when lock = '0' else resetp;  --hold interface in reset till DLL's lock.
---- XSASDRAM interface module          -------------
  uramctl : XSASDRAMJ
    generic map(                        --generic params for dram interface depend on board type and are defined above
      FREQ         => 100_000,
      CLK_DIV      => 2.0,
      PIPE_EN      => false,            --non-pipeline mode
      NROWS        => NROWS,            --4096->xsa50/100, 8192->xsa200/3S  
      NCOLS        => NCOLS )           --256->xsa50, 512->xsa100/200/3S  
    port map(
      ---- host side                    -----
      clk          => clk100,           -- 
      rst          => rst_int,          --internal reset waits for lock to switch to resetp
      clk1x        => synced50M,        --(clk/CLK_DIV) buffered and sync'ed 
      clk2x        => synced100M,       -- syncd 100Mhz
      rd           => hrd,              -- host-side SDRAM read control
      wr           => hwr,              -- host-side SDRAM write control
      lock         => lock,             -- valid synchronized clocks indicator
      rdPending    => open,             -- read still in pipeline
      opBegun      => opBegun,          -- memory read/write begun indicator
      earlyOpBegun => open,             -- memory read/write begun (async)
      rdDone       => open,             -- memory pipelined read done indicator
      done         => done,
      hAddr        => hAddr,            -- host-side address from project
      hDIn         => hdIn,             -- write data from project
      hDOut        => hdout,            -- SDRAM data output to project       
      ---- SDRAM side                   ----
      sclkfb       => sclkfb,           -- clock from SDRAM after PCB delays
      sclk         => sclk,             -- SDRAM clock sync'ed to master clock
      cke          => cke,              -- SDRAM clock enable
      cs_n         => cs_n,             -- SDRAM chip-select
      ras_n        => ras_n,            -- SDRAM RAS
      cas_n        => cas_n,            -- SDRAM CAS
      we_n         => we_n,             -- SDRAM write-enable
      ba           => ba,               -- SDRAM bank address
      saddr        => saddr,            -- SDRAM address
      sdata        => sdat,             -- SDRAM inout data bus
      dqmh         => dqmh,             -- SDRAM DQMH
      dqml         => dqml              -- SDRAM DQML
      );
---*********end interface to SDRAM control module****************

-----------------------------------------------------------------------
--SDRAM read/write state machine driven by SDRAM masterclk, compatible with
-- VBasic functions Load_MemUSB, Read_MemUSB
--The operation is controlled by the "mode" feature parameter sent by the PC program.
--This use of the mode parameter is an example of how the usb host can set a "feature"
-- in the FPGA, which is acting as the HID device. The mode byte can control how the
-- FPGA circuit operates, in much the same manner as setting dip switches on the XST board.
--The new_dat pulse tells the state machine that a new packet has arrived for loading.
--The rd_req pulse tells the state machine that the last read packet is being read by the PC
-----------------------------------------------------------------------
--Mem access state machine controls hrd, hwr, haddr, hdout, hdin, all declared in sdramj pkg.
  sm_clk <= synced50M;
  haddr  <= std_logic_vector(addr);     --host side address comes from rd/wr state machine

  MEM_OPS : process (rst_int, sm_clk, state, addr, opbegun) is
  begin
    if rst_int = '1' then
      state <= IDLE;                    --reset into the idle state to wait for a mode change.
      addr  <= (others => '0');
      hdin  <= (others => '0');
      hwr   <= '0';
      hrd   <= '0';

    elsif rising_edge(sm_clk) then      --
      case state is
        when IDLE                          =>  --we remain in idle state unless mode says to read or write mem.
          buf_ct                <= 0;   -- we always reset the packet array pointer just in case
          addr                  <= (others => '0');  --reset the dram address while we wait for a mode change.
          case mode(1 downto 0) is
            when "01"                      =>  --mode(1:0)=01, go wait for first packet from PC to write into dram    
              state             <= GETBYTES;  --
            when "10"                      =>  --mode(1:0)=10, go read the first packet from dram. 
              hrd               <= '1';  --start a packet read
              state             <= SENDBYTES;
            when others                    => null;
          end case;
                                        --------------------------------------------      
        when GETBYTES                      =>  --as long as project mode is 1, wait for the next usb OUT report.
          if mode(0) = '0' then         --check if memory write progam has completed.
            state               <= IDLE;  -- stop the write loop
          elsif new_dat = '1' then      --writes start when flag->1 indicating data in the buffer 
            buf_ct              <= 2;   --1st 2 bytes going out to dram
            hdin(7 downto 0)    <= PCOut(2);  --data low byte is latched.
            hdin(15 downto 8)   <= PCOut(1);  --data hi byte is latched
            hwr                 <= '1';  --first word ready, start writes
            state               <= WRITEP;
          end if;
        when WRITEP                        =>  --during writep states, we write the whole packet.
          if opBegun = YES then         -- sdram opbegun - stop write and latch next values.
            hwr                 <= '0';
            addr                <= addr + 1;  --latch next incremented addr, and latch the packet data        
            if buf_ct < 8 then
              buf_ct            <= buf_ct + 2;
              hdin(7 downto 0)  <= PCOut(2+buf_ct);  --data low byte is latched.
              hdin(15 downto 8) <= PCOut(1+buf_ct);  --data hi byte is latched
              state             <= WRPAUSE;
            else
              state             <= CHKFLG;  --finished this packet, wait for usb new_dat flag to go low.
                                            --hwr<= '0';  --turn off the DRAM write requests till next packet arrives
              buf_ct            <= 0;   --reset the mem in-out buffer pointer 
            end if;
          end if;
        when WRPAUSE                       =>  --start the next write
          hwr                   <= '1';
          state                 <= WRITEP;
        when CHKFLG                        =>
          if new_dat = '0' then
            state               <= GETBYTES;  --go back to get next pkt and continue writes   
          end if;
                                        ------------------------------------------      
        when SENDBYTES                     =>
                                        --During sendbytes state, we read a whole packet in from mem into the PCinput array. 
                                        --Then we wait for usb read request to pulse, indicating the packet was retrieved.
          if opbegun = YES then         --signal from sdram interface says read of addr has started
            hrd                 <= '0';  --stop the read
            addr                <= addr + 1;  --latch next incremented addr. 
          elsif done = YES then         --latch the dram data when the dram read completes
            mempacket(2+buf_ct) <= hdout(7 downto 0);  --data low byte is latched.
            mempacket(1+buf_ct) <= hdout(15 downto 8);  --data hi byte is latched
            if buf_ct < 6 then          --we read and latch 4 times, for buf_ct= 0,2,4,6
              hrd               <= '1';  --if the packet buffer not full, keep the reads going
              buf_ct            <= buf_ct + 2;
              state             <= SENDBYTES;  --continue reading dram to fill the packet
            else
              state             <= RDPAUSE1;  --finished this packet, wait for usb rd_req pulse.
                                              --hrd<= '0';  --turn off the DRAM read requests
              buf_ct            <= 0;   --reset the mempacket buffer pointer
            end if;
          end if;
                                        --We pause here for the packet just read from DRAM, to be retrieved by the host PC.     
        when RDPAUSE1                      =>  --wait till usb rd_req goes high-low to loop back to read next packet.
          if mode(1) = '0' then         --chk bit1 - The reading program has terminated if we leave mode 2.
            state               <= IDLE;
          elsif rd_req = '1' then
            state               <= RDPAUSE2;  --wait for USB rd_req to complete the toggle indicating packet sent     
          end if;
        when RDPAUSE2                      =>  --wait till rd_req goes back low to loop back for next packet
          if rd_req = '0' then
            hrd                 <= '1';  --start next read
            state               <= SENDBYTES;  --go load next packet. addr is pointing to next loc to be read    
          end if;

        when others => null;
      end case;
    end if;
  end process MEM_OPS;
------------------------------------------------------------

end arch;
